;-------------------------------------------------------------------------------
; Init pspad lib functions
;
;   Copyright (C) 2008, 2009 Alesc
;   This program is free software: you can redistribute it and/or modify
;   it under the terms of the GNU General Public License.
;-------------------------------------------------------------------------------

;---------------------------------------
; InitPSPlib
;
; Arguments: action_key - shortcut of the pspad jscript
;            psp_jscript_name - jscript name without extension
;
; Returns:
;   HOT_OK / HOT_ERR
;
; Description:
;   Initializes psp_lib. Must be called before
;   any other psp function is called.
;
; Important globals initialized:
;   PSP_ACTION_KEY      shortcut to pspad jscript
;   PSP_EXE             pspad executable with full path
;   PSP_PATH            main pspad path
;   PSP_JSCRIPTS_PATH   jscripts path
;   PSP_ACTION_FILE     jscript input file
;   PSP_OUT_FILE        jscript output file
;   PSP_JSCRIPT         jscript name with full path
;   PSP_VER             pspad version
;   PSP_BUILD_VER       pspad build number
;---------------------------------------
InitPSPlib(action_key, psp_jscript_name, prog_tmp_path)
{
    global

    DbgIn(A_ThisFunc)

    if (action_key = "" or psp_jscript_name = "")
    {
        Report("Error", "Can not initalize psp lib. Invalid arguments.")
        DbgOut(A_ThisFunc)
        return %HOT_ERR%
    }

    ; call pspad
    PSP_ACTION_KEY := action_key

    if GetPSPPaths() = HOT_ERR
    {
        DbgOut(A_ThisFunc)
        return %HOT_ERR%
    }

    PSP_ACTION_FILE = %PSP_JSCRIPTS_PATH%psplib_action.ahk2psp
    PSP_OUT_FILE = %PSP_JSCRIPTS_PATH%psplib_output.psp2ahk
    DbgPrint("Action file [" . PSP_ACTION_FILE . "].")
    DbgPrint("Output file [" . PSP_OUT_FILE . "].")

    ; update pspad jscript with which ahk is going to communicate.
    if UpdatePSPScript(psp_jscript_name, prog_tmp_path) = HOT_ERR
    {
        DbgOut(A_ThisFunc)
        return %HOT_ERR%
    }

    DbgOut(A_ThisFunc)
    return %HOT_OK%
}
;---------------------------------------
; GetPSPPaths
;
; Arguments: none
;
; Returns:
;   HOT_OK / HOT_ERR
;
; Description:
;   Reads pspad install location from
;   win.registry and sets globals.
;
; Important globals initialized:
;   PSP_EXE
;   PSP_PATH
;   PSP_JSCRIPTS_PATH
;---------------------------------------
GetPSPPaths()
{
    global

    DbgIn(A_ThisFunc)

    ; get/set paths to pspad editor
    RegRead, PSP_EXE, HKEY_CURRENT_USER, Software\PSPad, PSPadPath
    if ErrorLevel
    {
        Report("Error", "Can not read PSPad path from windows registry.(HKEY_CURRENT_USER\Software\PSPad)")
        DbgOut(A_ThisFunc)
        return %HOT_ERR%
    }

    SplitPath, PSP_EXE,, PSP_PATH
    ; sometimes PSP_EXE is in doublequotes
    if RegExMatch(PSP_PATH, "^""")
      StringTrimLeft, PSP_PATH, PSP_PATH, 1
    IfNotExist, %PSP_PATH%
    {
        Report("Error", "Can not find PSPad path [" . PSP_PATH . "].")
        DbgOut(A_ThisFunc)
        return %HOT_ERR%
    }

    PSP_JSCRIPTS_PATH = %PSP_PATH%\Script\JScript\
    IfNotExist, %PSP_JSCRIPTS_PATH%
    {
        Report("Error", "Can not find PSPad scripts path [" . PSP_JSCRIPTS_PATH . "].")
        DbgOut(A_ThisFunc)
        return %HOT_ERR%
    }

    DbgOut(A_ThisFunc)
    return %HOT_OK%
}
;---------------------------------------
; UpdatePSPScript
;
; Arguments: psp_jscript_name - jscript name without extension
;
; Returns:
;   HOT_OK / HOT_ERR
;
; Description:
;   Updates pspad jscript with new communication
;   parameters only if modified.
;---------------------------------------
UpdatePSPScript(psp_jscript_name, prog_tmp_path)
{
    global PSP_JSCRIPT, PSP_JSCRIPTS_PATH, PSP_ACTION_KEY
    global PSP_ACTION_FILE, PSP_OUT_FILE, PSP_VER, PSP_BUILD_VER
    global HOT_OK, HOT_ERR

    DbgIn(A_ThisFunc)

    PSP_JSCRIPT = %PSP_JSCRIPTS_PATH%%psp_jscript_name%.js
    IfNotExist, %PSP_JSCRIPT%
    {
        Report("Error", "Can not find PSPad jscript.[" . PSP_JSCRIPT . "].")
        DbgOut(A_ThisFunc)
        return %HOT_ERR%
    }

    ; prepare strings for jscript modification
    actionkey_start_str = addMenuItem("%psp_jscript_name%", "%psp_jscript_name%", "%psp_jscript_name%", "
    actionkey_end_str = ");
    actionfile_start_str = var psp_action_file = "
    outfile_start_str = var psp_out_file = "
    vars_end_str = ";

    convActionKey := AhkToPSPHotkey(PSP_ACTION_KEY)

    ; create tmp jscript to work on
    if (tmp_psp_jscript := CreateTmpFile(prog_tmp_path, "psplib" . %psp_jscript_name%)) = ""
        goto exit_UpdatePSPScript

    file_modified = 0
    ; jscript re-write loop
    Loop, read, %PSP_JSCRIPT%
    {
        ; replace action key line
        ; TODO need to retrieve previous ACTION_KEY because of GetPSPVersion below
        if retval := ReplaceJSLine(actionkey_start_str, convActionKey, actionkey_end_str, previousKey, tmp_psp_jscript) = HOT_OK
        {
            file_modified = 1
            continue
        }
        else if retval = HOT_ERR
            goto exit_UpdatePSPScript

        if retval := FileAppendEx(A_LoopReadLine . "`r`n", 0, tmp_psp_jscript) = HOT_ERR
            goto exit_UpdatePSPScript
    }

    tmp_action_key = %PSP_ACTION_KEY%
    PSP_ACTION_KEY := PSPHotkeyToAhk(previousKey)
    ; pspad version format: 4.5.4 (2331)
    psp_ver_str := ExecPSPFunc("hots_core.writePspadVersion")
    StringSplit, psp_ver_arr, psp_ver_str, .(),
    PSP_VER = %psp_ver_arr1%%psp_ver_arr2%%psp_ver_arr3%
    PSP_BUILD_VER = %psp_ver_arr4%
    PSP_ACTION_KEY = %tmp_action_key%

    ; if jscript was modified overwrite it
    if file_modified = 1
    {
        FileMove, %tmp_psp_jscript%, %PSP_JSCRIPT%, 1
        ; pspad ver 454 2331 has new recompile scripts action
        if (PSP_VER >= 454 and PSP_BUILD_VER >= 2331)
        {
            ExecPSPFunc("runPSPadAction(aScriptRecompile)")
        }
        else
        {
            SendToPSP("!T", 1, 1)
            SendToPSP("{Right}", 1, 1)
            SendToPSP("{Enter}", 1, 1)
        }
    }

exit_UpdatePSPScript:
    ; ignore errors...file may not exist
    DeleteTmpFiles(prog_tmp_path, "psplib")

    DbgOut(A_ThisFunc)
    return %retval%
}
;---------------------------------------
; ReplaceJSLine
;
; Arguments: search_str - search and start string
;            insert_str - string to insert in the midle
;            end_str - end string
;
; Returns:
;   HOT_OK / HOT_ERR / NOT_MODIFIED
;
; Description:
;   Searches and replaces if needed A_LoopReadLine
;   with Search_strInsert_strEnd_str + comments to
;   target_file. HOT_OK for line was replaced.
;---------------------------------------
ReplaceJSLine(search_str, insert_str, end_str, ByRef replaced_str, target_file)
{
    global HOT_OK, HOT_ERR

    DbgIn(A_ThisFunc)

    NOT_MODIFIED = -1
    psplib_comment = // modified on %A_Now%(YYYYMMDDHH24MISS) by %A_ScriptName%

    IfInString, A_LoopReadLine, %search_str%
    {
        ; addMenuItem("hots", "hots", "hots", "Ctrl+Alt+G");
        ;MsgBox %A_LoopReadLine%
        position := RegExMatch(A_LoopReadLine, "^.+""(.+)""\);.+$", line)
        if (position != 0)
        {
            replaced_str = %line1%
        }
        linetmp = %search_str%%insert_str%%end_str%
        IfInString, A_LoopReadLine, %linetmp%
        {
            DbgOut(A_ThisFunc)
            return %NOT_MODIFIED%
        }

        linetmp = %linetmp%%psplib_comment%
        if FileAppendEx(linetmp . "`r`n", 0, target_file) = HOT_ERR
        {
            DbgOut(A_ThisFunc)
            return %HOT_ERR%
        }

        ; file modified
        DbgOut(A_ThisFunc)
        return %HOT_OK%
    }

    DbgOut(A_ThisFunc)
    return %NOT_MODIFIED%
}
;-------------------------------------------------------------------------------
; pspad help functions
;-------------------------------------------------------------------------------

;---------------------------------------
; SendToPSP
;
; Arguments: cmd - keystrokes/mouse clicks to send
;            isCmd - it is strings0 or commmand1
;
; Returns:
;   nothing
;
; Description:
;   Sends keystrokes or mouse clicks to pspad.
;   PSPad is started if not running or 
;   window activated if not active. 
;   If command is sent(isCmd=1) than lower the chars.
;---------------------------------------
SendToPSP(cmd, isCmd, runPSP)
{
    global PSP_EXE

    DbgIn(A_ThisFunc)

    IfWinExist, PSPad
    {
        DbgPrint("PSPad running. Activate window.")
        WinActivate
    }
    else if runPSP = 1
    {
        DbgPrint(PSP_EXE . " about to run.")
        Run, %PSP_EXE%
        WinWaitActive, PSPad
    }

    ; checking again just to be safe
    IfWinActive, PSPad
    {
        if isCmd = 1
            StringLower, cmd, cmd
        DbgPrint("Sending " . cmd)
        SendInput, %cmd%
    }
    else if runPSP = 1
    {
        Report("Error", "PSPad not running or can not run.")
        DbgPrint("PSPad not running, window not active or can not run.")
    }

    DbgOut(A_ThisFunc)
}
;---------------------------------------
; SendToPSPWait
;
; Arguments: cmd - command to send
;            out - output filename
;
; Returns:
;   Hotkey
;
; Description:
;   Command is sent to PSPad and waits for
;   output file reply.
;---------------------------------------
SendToPSPWait(cmd, out)
{
    global HOT_OK, HOT_ERR

    DbgIn(A_ThisFunc)

    ; just to be safe
    FileDelete, %out%
    SendToPSP(cmd, 1, 1)

    if WaitForFile(out) = HOT_ERR
    {
        DbgPrint("No output file " . out . " from PSPad.")
        DbgOut(A_ThisFunc)
        return %HOT_ERR%
    }

    DbgOut(A_ThisFunc)
    return %HOT_OK%
}
;---------------------------------------
; AhkToPSPHotkey
;
; Arguments: ahk - hotkey
;
; Returns:
;   Hotkey
;
; Description:
;   Converts hotkey to pspad jscript friendly
;   format.
;---------------------------------------
AhkToPSPHotkey(ahk)
{
    ahk := RegExReplace(ahk, "\+", "Shift+")
    ahk := RegExReplace(ahk, "!", "Alt+")
    ahk := RegExReplace(ahk, "\^", "Ctrl+")
    ahk := RegExReplace(ahk, "#", "Win+")

    return %ahk%
}
PSPHotkeyToAhk(psp)
{
    psp := RegExReplace(psp, "(Shift\+|Shift)", "+")
    psp := RegExReplace(psp, "(Alt\+|Alt)", "!")
    psp := RegExReplace(psp, "(Ctrl\+|Ctrl)", "^")
    psp := RegExReplace(psp, "(Win\+|Win)", "#")

    return %psp%
}
;---------------------------------------
; AhkToPSPPath
;
; Arguments: path - path/directory
;
; Returns:
;   path
;
; Description:
;   Converts path to pspad jscripts friendly
;   format. Path in jscript must have double
;   backslash.
;---------------------------------------
AhkToPSPPath(path)
{
    path := RegExReplace(path, "\\", "\\")
    return %path%
}
;---------------------------------------
; ExecPSPFunc
;
; Arguments: function - function to execute
;
; Returns:
;   function output
;
; Description:
;   
;---------------------------------------
ExecPSPFunc(function)
{
    global
    local output, level

    DbgIn(A_ThisFunc)

    DbgPrint("Sending [" . function . "] to PSPad.")
    DeletePSPConn()

    if FileAppendEx(function . "`r`n", "1", PSP_ACTION_FILE) = HOT_ERR
        goto exit_ExecPSPFunc

    if SendToPSPWait(PSP_ACTION_KEY, PSP_OUT_FILE) = HOT_ERR
        goto exit_ExecPSPFunc

    FileReadLine, output, %PSP_OUT_FILE%, 1
    if ErrorLevel
        Report("Error", "Output file '" . PSP_OUT_FILE . "' read error.")

    if (level := ParsePSPError(output)) = "Normal"
        goto exit_ExecPSPFunc

    Report(level, output)
    output =

exit_ExecPSPFunc:
    DeletePSPConn()

    DbgOut(A_ThisFunc)
    return %output%
}
ParsePSPError(byref output)
{
    level = Normal
    position := RegExMatch(output, "HOTSERR:(.+): .+", line)
    if (position = 0)
        return %level%

    return line1
}
DeletePSPConn()
{
    global

    IfExist, %PSP_ACTION_FILE%
    {
        FileDelete, %PSP_ACTION_FILE%
        if ErrorLevel
            Report("Warning", "Faild to delete conn.file [" . PSP_ACTION_FILE . "].")
    }
    IfExist, %PSP_OUT_FILE%
    {
        FileDelete, %PSP_OUT_FILE%
        if ErrorLevel
            Report("Warning", "Failed to delete conn.file [" . PSP_OUT_FILE . "].")
    }
}
GetPSPCaretPos(byref filename, byref ycaret, byref xcaret)
{
    global HOT_OK, HOT_ERR

    output := ExecPSPFunc("hots_core.writeCaretPosition")
    if !RegExMatch(output, "Filename=(.+), Y=(.+), X=(.+)", line)
    {
        Report("Error", "Improper format of the caret position line [" . output . "].")
        return %HOT_ERR%
    }

    filename = %line1%
    ycaret = %line2%
    xcaret = %line3%
    return %HOT_OK%
}
